#ifndef QP_FRAME_H
#define QP_FRAME_H

#include "qp_util.h"

#include <ffmpeg/avcodec.h>
#include <ffmpeg/avformat.h>

/* 
 * Basic frame manipulation routines. This should remain general enough to
 * be used in ffmpeg-php or rmovie. 
 */

typedef struct {
    AVFrame *av_frame;
    int width;
    int height;
    int pixel_format;
    int keyframe;
    int64_t pts;
} qp_frame_context;

int qp_get_frame_height(qp_frame_context *frame);
int qp_get_frame_width(qp_frame_context *frame);
const char* qp_get_frame_pixel_format_name(qp_frame_context *frame_ctx);

/* This should be static */
void qp_free_av_frame(AVFrame *av_frame);

qp_frame_context* qp_alloc_frame_ctx(qp_malloc_t malloc_func);
void qp_free_frame_ctx(qp_frame_context *frame_ctx, qp_free_t free_func);


int qp_convert_av_frame(AVFrame **av_frame, int pixel_format, 
        int new_format, int width, int height, int free_buffers);

int qp_convert_frame(qp_frame_context *frame, int new_fmt);
int qp_crop_frame(qp_frame_context *frame,
        int crop_top, int crop_bottom, int crop_left, int crop_right);
int qp_resample_frame(qp_frame_context *frame,
        int wanted_width, int wanted_height, int crop_top, int crop_bottom,
        int crop_left, int crop_right);

int qp_output_frame_as_string(qp_frame_context *frame, unsigned char**);

double qp_get_pts(qp_frame_context *frame);
long qp_frame_is_keyframe(qp_frame_context *frame);

#endif
