#ifndef QP_MOVIE_H
#define QP_MOVIE_H

#include <ffmpeg/avcodec.h>
#include <ffmpeg/avformat.h>

#include "qp_util.h"
#include "qp_frame.h"

#define QP_FRAME_BASED 0
#define QP_TIME_BASED 1

typedef struct {
    AVFormatContext *fmt_ctx;
    AVCodecContext *codec_ctx[MAX_STREAMS];
    int64_t last_pts;
    int frame_number; /* only valid in time based movies */
    int mode; 
    offset_t last_file_pos;

    /* qp_export_movie context stuff */

    /* Video buffer */
    uint8_t *video_outbuf;
    int video_outbuf_size;

    /* Audio buffer */
    int16_t *audio_samples;
    unsigned int audio_samples_size;

    uint8_t *audio_outbuf;
    int audio_outbuf_size;
    int audio_input_frame_size;

    /* User defined. */
    void* user_defined;

    /* Hack for php until I fix it up to use 'user_defined' */
    long udef_long;
} qp_movie_context;

void qp_free_movie_ctx(qp_movie_context *movie_ctx, qp_free_t free_func);
qp_movie_context* qp_alloc_movie_ctx();
int qp_open_movie_file(qp_movie_context *movie_ctx, const char* filename);

float qp_get_frame_rate(qp_movie_context *movie_ctx);
float qp_get_duration(qp_movie_context *movie_ctx);
long qp_get_frame_count(qp_movie_context *movie_ctx);
char* qp_get_file_name(qp_movie_context *movie_ctx);

/* mp3 metadata accessors */
char* qp_get_comment(qp_movie_context *movie_ctx);
char* qp_get_title(qp_movie_context *movie_ctx);
char* qp_get_author(qp_movie_context *movie_ctx);
char* qp_get_copyright(qp_movie_context *movie_ctx);
char* qp_get_album(qp_movie_context *movie_ctx);
char* qp_get_genre(qp_movie_context *movie_ctx);
int qp_get_year(qp_movie_context *movie_ctx);
int qp_get_track(qp_movie_context *movie_ctx);

int qp_get_movie_width(qp_movie_context *movie_ctx);
int qp_get_movie_height(qp_movie_context *movie_ctx);
long qp_get_current_frame_number(qp_movie_context *movie_ctx);
const char* qp_get_movie_pixel_format_name(qp_movie_context *movie_ctx);
int qp_get_bit_rate(qp_movie_context *movie_ctx);
const char* qp_get_codec_name(qp_movie_context *movie_ctx, int type);
int qp_has_audio(qp_movie_context *movie_ctx);
int qp_has_video(qp_movie_context *movie_ctx);
int qp_get_num_audio_channels(qp_movie_context *movie_ctx);
double qp_get_pixel_aspect_ratio(qp_movie_context *movie_ctx);
int qp_export_movie(qp_movie_context *movie_ctx, const char *filename);
int qp_get_codec_sample_rate(qp_movie_context *movie_ctx, int type);
int qp_get_codec_bit_rate(qp_movie_context *movie_ctx, int type);


#define QP_GETFRAME_KEYFRAME -1
#define QP_GETFRAME_NEXTFRAME 0
int qp_get_frame(qp_movie_context *movie_ctx, int wanted_frame,
        qp_frame_context **frame_ctx_handle);

#endif /* QP_MOVIE_H */
